package tk.to_p.blankedoutbox.garlicclone.ai;

import tk.to_p.blankedoutbox.garlicclone.core.GamePlay;
import tk.to_p.blankedoutbox.garlicclone.core.GarlicConst;
import tk.to_p.blankedoutbox.garlicclone.core.Player;
import tk.to_p.blankedoutbox.garlicclone.core.TetriMinos;

/**
 * This class tries to fix some of T-Spin bugs of original Flash version of TetrisAI.
 */
public class FixedTetrisAI extends TetrisAI {
	/** Current combo count */
	public int curComboCnt_;

	public FixedTetrisAI() {
		super();
	}
	public FixedTetrisAI(GamePlay gamePlay) {
		super(gamePlay);
	}

	@Override
	public void Run() {
		// It seems the original AI code wanted to store the combo count to somewhere, so...
		curComboCnt_ = gamePlay_.GetStatistics().curComboCnt_;

		if(curState_ == STATE_THINK_START)
		{
			Player.Piece curPiece = gamePlay_.GetPlayer().GetCurPiece();

			if(curPiece != null) {
				PrepareThink();
				curMovingPieceId_ = curPiece.id_;
				curState_ = STATE_THINKING;
				curThinkDelayFrame_ = reqThinkDelayFrame_;
			}
		}
		if(curState_ == STATE_THINKING)
		{
			if(--curThinkDelayFrame_ <= 0) {
				gamePlay_.SetControllerFlags(0);
				Player.Piece curPiece = gamePlay_.GetPlayer().GetCurPiece();

				if(curPiece != null) {
					if (curMovingPieceId_ != curPiece.id_) {
						curState_ = STATE_THINK_START;
					} else if(Think()) {
						curState_ = STATE_MOVING;
						curMoveDelayFrame_ = reqMoveDelayFrame_;
					}
				}
				return;
			}
		}
		if(curState_ == STATE_MOVING)
		{
			int HOLDACTION_NONE = 0;
			int HOLDACTION_EXECUTE = 1;
			//int HOLDACTION_UNUSED = 2;

			if(curMoveDelayFrame_-- > 0) {
				gamePlay_.SetControllerFlags(0);
			} else {
				curMoveDelayFrame_ = reqMoveDelayFrame_;
				int ctrlFlag = 0;
				Player.Piece curPiece = gamePlay_.GetPlayer().GetCurPiece();

				if(curPiece != null)
				{
					if(bestMove_.hold == HOLDACTION_EXECUTE) {
						if (curMovingPieceId_ != curPiece.id_) {
							bestMove_.hold = HOLDACTION_NONE;
							curMovingPieceId_ = curPiece.id_;
						} else {
							ctrlFlag = GarlicConst.ControllerFlag_HOLD;
						}
					} else {
						if(curMovingPieceId_ != curPiece.id_) {
							curState_ = STATE_THINK_START;
						} else {
							int curX;
							int curY;
							curX = curPiece.x_;
							curY = curPiece.y_;

							int xn;
							int an;
							if(curY < bestMove_.y0) {
								xn = bestMove_.x0;
								an = bestMove_.a0;
							} else {
								xn = bestMove_.x1;
								an = bestMove_.a1;
							}

							if (curX < xn) {
								ctrlFlag |= GarlicConst.ControllerFlag_RIGHT;
							} else if (curX > xn) {
								ctrlFlag |= GarlicConst.ControllerFlag_LEFT;
							}

							if (an != curPiece.face_) {
								int face = an - curPiece.face_;
								if(face < 0) face += GarlicConst.FaceType_Count;

								if(face > GarlicConst.FaceType_Count / 2) {
									ctrlFlag |= GarlicConst.ControllerFlag_CCW;
								} else {
									ctrlFlag |= GarlicConst.ControllerFlag_CW;
								}
							}

							if (bestMove_.a0 == bestMove_.a1
							 && bestMove_.x0 == bestMove_.x1
							 && curX == xn && an == curPiece.face_)
							{
								ctrlFlag |= GarlicConst.ControllerFlag_HARD;
								curState_ = STATE_THINK_START;
							}
							// Added for T-Spin instant locks
							else if (curY >= bestMove_.y0 && an == curPiece.face_)
							{
								ctrlFlag |= GarlicConst.ControllerFlag_HARD;
							}
							else if (curX == xn && an == curPiece.face_)
							{
								ctrlFlag |= GarlicConst.ControllerFlag_SOFT;
							}
						}
					}
				} /* if(curPiece != null) */

				gamePlay_.SetControllerFlags(ctrlFlag);
				if(curState_ == STATE_MOVING) {
					curMoveDelayFrame_ = Math.max(1, curMoveDelayFrame_);
				}
			}
		} /* if(curState_ == STATE_MOVING) */
	}

	@Override
	public boolean DoThink(MoveData moveData) {
		ThinkNextBlockResult thinkResult = new ThinkNextBlockResult();
		boolean curPieceIsT = (nextBlock_.get(CUR_BLOCK) == GarlicConst.TetriminoType_T);
		boolean nextPieceIsT = (nextBlock_.get(CUR_BLOCK + 1) == GarlicConst.TetriminoType_T);
		boolean holdIsT = canHold_ && (nextBlock_.get(HOLD_BLOCK) == GarlicConst.TetriminoType_T);
		boolean isHold = false;
		ThinkNextBlockResult thinkResultNext = new ThinkNextBlockResult();
		ThinkNextBlockResult thinkResultHold = new ThinkNextBlockResult();

		ThinkNextBlock(CUR_BLOCK, thinkResultNext);
		if(canHold_) ThinkNextBlock(HOLD_BLOCK, thinkResultHold);

		if(thinkResultHold.best > thinkResultNext.best + holdMinDiff_) {
			thinkResult = thinkResultHold;
			isHold = true;
		} else {
			thinkResult = thinkResultNext;
		}

		if(enableTSpin_) {
			if(curPieceIsT || holdIsT) {
				MoveData tSpinPosResult = new MoveData();
				boolean isTSpinPosFound = FindTSpinPos(BLOCK_MIN_LEFT, width_, tSpinPosResult);

				if(holdIsT && isTSpinPosFound) {
					MakeTSpinMatrixRetObject tMatrixResult = new MakeTSpinMatrixRetObject();
					if(MakeTSpinMatrix(CUR_BLOCK, true, tMatrixResult) && tMatrixResult.best > tSpinPosResult.best) {
						moveData.x1 = tMatrixResult.x;
						moveData.x0 = moveData.x1;
						moveData.y1 = tMatrixResult.y;
						moveData.y0 = moveData.y1;
						moveData.a1 = tMatrixResult.a;
						moveData.a0 = moveData.a1;
						moveData.hold = 0;
						return true;
					}
				}

				if(canHold_ && curPieceIsT && isTSpinPosFound) {
					MakeTSpinMatrixRetObject tMatrixResult = new MakeTSpinMatrixRetObject();
					if(MakeTSpinMatrix(HOLD_BLOCK, true, tMatrixResult) && tMatrixResult.best > tSpinPosResult.best) {
						moveData.x1 = tMatrixResult.x;
						moveData.x0 = moveData.x1;
						moveData.y1 = tMatrixResult.y;
						moveData.y0 = moveData.y1;
						moveData.a1 = tMatrixResult.a;
						moveData.a0 = moveData.a1;
						moveData.hold = 1;
						return true;
					}
				}

				if (isTSpinPosFound && tspinMinDiff_ + tSpinPosResult.best > thinkResult.best) {
					moveData.x0 = tSpinPosResult.x0;
					moveData.x1 = tSpinPosResult.x1;
					moveData.y0 = tSpinPosResult.y0;
					moveData.y1 = tSpinPosResult.y1;
					moveData.a0 = tSpinPosResult.a0;
					moveData.a1 = tSpinPosResult.a1;
					moveData.hold = (!curPieceIsT && holdIsT) ? 1 : 0;
					return true;
				}
			}

			if(nextPieceIsT || holdIsT) {
				MakeTSpinMatrixRetObject info = new MakeTSpinMatrixRetObject();
				if(MakeTSpinMatrix(CUR_BLOCK, false, info) && info.best > thinkResult.best) {
					moveData.x1 = info.x;
					moveData.x0 = moveData.x1;
					moveData.y1 = info.y;
					moveData.y0 = moveData.y1;
					moveData.a1 = info.a;
					moveData.a0 = moveData.a1;
					moveData.hold = 0;
					return true;
				}
			}

			if(canHold_ && !isHold && curPieceIsT && !holdIsT && thinkResultHold.best + holdMinDiff_ > thinkResultNext.best) {
				thinkResult = thinkResultHold;
				isHold = true;
			}
		}

		moveData.x0 = thinkResult.xBest;
		moveData.x1 = thinkResult.xBest;
		moveData.y0 = thinkResult.yBest;
		moveData.y1 = thinkResult.yBest;
		moveData.a0 = thinkResult.aBest;
		moveData.a1 = thinkResult.aBest;
		moveData.hold = (isHold ? 1 : 0);

		return true;
		//return false;
	}

	@Override
	public boolean FindTSpinPos(int xMin, int xMax, MoveData objReturn) {
		objReturn.best = EVAL_MIN;
		boolean result = false;

		for(int face = 0; face < 4; face++) {
			if(face == 2) {
				// Do Nothing
			} else {
				TetriMinos.TetriMino tetriMino = GarlicConst.TetriMinos_[GarlicConst.TetriminoType_T].tetriMinos_[face];

				for(int x = xMin; x < xMax; x++) {
					int y = height_ / 2;
					for(;;) {
						if(!(TestShape(x, y, tetriMino) && y < height_)) break;
						y++;
					}
					y--;

					if(TestShape(x, y, tetriMino)) {
						int pattern = GetMatrixPattern(x - 1, y, 5, 3);
						IsTSpinPatternRetObject tPatternResult = new IsTSpinPatternRetObject();

						if(IsTSpinPattern(pattern, tPatternResult) && (tPatternResult.a0 == face)) {
							int x2 = x + tPatternResult.xDiff;
							int y2 = y + tPatternResult.yDiff;
							ToggleShape(x2, y2, GarlicConst.TetriMinos_[GarlicConst.TetriminoType_T].tetriMinos_[tPatternResult.a1]);

							int cornerCount = 0;
							if(GetField(x2    , y2    )) cornerCount++;
							if(GetField(x2 + 2, y2    )) cornerCount++;
							if(GetField(x2    , y2 + 2)) cornerCount++;
							if(GetField(x2 + 2, y2 + 2)) cornerCount++;

							boolean isProperSpin = false;
							if (GarlicConst.FaceType_North == face && GetField(x2 + 1, y2 + 2) ||
								GarlicConst.FaceType_East  == face && GetField(x2    , y2 + 1) ||
								GarlicConst.FaceType_South == face && GetField(x2 + 1, y2    ) ||
								GarlicConst.FaceType_West  == face && GetField(x2 + 2, y2 + 1))
							{
								isProperSpin = true;
							}

							if(cornerCount >= 3) {
								int noHoleCount = 0;

								for(int y3 = 0; y3 < 3; y3++) {
									if(y2 + y3 >= height_) break;

									boolean isNoHoles = true;
									for(int x3 = 0; x3 < width_; x3++) {
										if(!GetField(x3, y2 + y3)) {
											isNoHoles = false;
											break;
										}
									}

									if(isNoHoles) noHoleCount++;
								}

								if(noHoleCount > 0) {
									int tspinMinDiff = tspinMinDiff_;
									int points = Evalfield(0) + (noHoleCount - (isProperSpin ? 1 : 0)) * tspinMinDiff;
									if(points > objReturn.best) {
										objReturn.best = points;
										objReturn.x0 = x;
										objReturn.x1 = x;
										objReturn.y0 = y;
										objReturn.y1 = y;
										objReturn.a0 = tPatternResult.a0;
										objReturn.a1 = tPatternResult.a1;
										result = true;
									}
								}
							}

							ToggleShape(x2, y2, GarlicConst.TetriMinos_[GarlicConst.TetriminoType_T].tetriMinos_[tPatternResult.a1]);
						}
					}
				}
			}
		}

		return result;
	}

	@Override
	public boolean MakeTSpinMatrix(int curDepth, boolean checkAll, MakeTSpinMatrixRetObject objReturn) {
		objReturn.best = EVAL_MIN;
		boolean result = false;

		for(int i = 0; i < 4; i++) {
			TetriMinos.TetriMino tetriMino = GarlicConst.TetriMinos_[nextBlock_.get(curDepth)].tetriMinos_[i];

			for(int x = BLOCK_MIN_LEFT; x < width_; x++) {
				int y = height_ / 2;
				for(;;) {
					if(!(TestShape(x, y, tetriMino) && y < height_)) break;
					y++;
				}
				y--;

				if(TestShape(x, y, tetriMino)) {
					ToggleShape(x, y, tetriMino);
					MoveData tSpinPosResult = new MoveData();
					boolean tSpinPosSuccess = false;
					if(checkAll) {
						tSpinPosSuccess = FindTSpinPos(BLOCK_MIN_LEFT, width_, tSpinPosResult);
					} else {
						// Those tetriMino.xMin_ and tetriMino.xMax_ didn't have a underbar at the end
						tSpinPosSuccess = FindTSpinPos(x + tetriMino.xMin_ + BLOCK_MIN_LEFT, x + tetriMino.xMax_, tSpinPosResult);
					}
					ToggleShape(x, y, tetriMino);

					if(tSpinPosSuccess) {
						if (tSpinPosResult.best > objReturn.best) {
							objReturn.best = tSpinPosResult.best;
							objReturn.a = i;
							objReturn.x = x;
							objReturn.y = y;
							result = true;
						}
					}
				}
			}
		}

		return result;
	}
}
