package tk.to_p.blankedoutbox.garlicclone.ai;


import java.io.FileInputStream;
import java.io.IOException;

import se.krka.kahlua.converter.LuaConverterManager;
import se.krka.kahlua.converter.LuaNumberConverter;
import se.krka.kahlua.integration.LuaCaller;
import se.krka.kahlua.integration.LuaReturn;
import se.krka.kahlua.integration.annotations.LuaClass;
import se.krka.kahlua.integration.annotations.LuaMethod;
import se.krka.kahlua.integration.expose.LuaJavaClassExposer;
import se.krka.kahlua.j2se.J2SEPlatform;
import se.krka.kahlua.luaj.compiler.LuaCompiler;
import se.krka.kahlua.vm.KahluaTable;
import se.krka.kahlua.vm.KahluaThread;
import se.krka.kahlua.vm.LuaClosure;
import tk.to_p.blankedoutbox.garlicclone.core.GamePlay;

/**
 * This AI will load a TOJ's Lua script and execute it!
 */
@LuaClass
public class LuaTetrisAI extends FixedTetrisAI {
	public LuaConverterManager converterManager;
	public J2SEPlatform platform;
	public KahluaTable env;
	public KahluaThread thread;
	public LuaCaller caller;
	public LuaJavaClassExposer exposer;
	public LuaClosure closure;
	public String strLuaScript = "res/ai/test_tspin.lua";

	public LuaTetrisAI() {
		super();
	}
	public LuaTetrisAI(GamePlay gamePlay) {
		super(gamePlay);
	}
	public LuaTetrisAI(String filename) {
		this.strLuaScript = filename;
	}
	public LuaTetrisAI(GamePlay gamePlay, String filename) {
		this.strLuaScript = filename;
		Initialize(gamePlay);
	}

	public void showLuaError(LuaReturn ret) {
		if(!ret.isSuccess()) {
			System.err.println(ret.getErrorString() + ret.getLuaStackTrace());
		}
	}

	@Override
	public boolean DoInitialize() {
		converterManager = new LuaConverterManager();
		LuaNumberConverter.install(converterManager);
		platform = new J2SEPlatform();
		env = platform.newEnvironment();
		thread = new KahluaThread(platform, env);
		caller = new LuaCaller(converterManager);
		exposer = new LuaJavaClassExposer(converterManager, platform, env);
		exposer.exposeGlobalFunctions(this);

		try {
			System.out.println("Loading Lua script from " + strLuaScript + "...");

			FileInputStream in = new FileInputStream(strLuaScript);
			closure = LuaCompiler.loadis(in, "", env);
			in.close();

			LuaReturn ret = caller.protectedCall(thread, closure);
			if(!ret.isSuccess()) {
				System.err.println("This Lua script has problem(s):");
				showLuaError(ret);
				return false;
			}

			Object objStart = closure.env.rawget("Start");
			if((objStart != null) && (objStart instanceof LuaClosure)) {
				LuaReturn retStart = caller.protectedCall(thread, objStart);
				if(!retStart.isSuccess()) {
					System.err.println("Start() function failed:");
					showLuaError(retStart);
					return false;
				} else {
					System.out.println("Start() OK");
				}
			}

			System.out.println("Load successful");
		} catch (IOException e) {
			System.err.println("Lua script '" + strLuaScript + "' load failed");
			e.printStackTrace();
			return false;
		}

		return true;
	}

	@Override
	public void DoShutdown() {
		Object objEnd = closure.env.rawget("End");
		if((objEnd != null) && (objEnd instanceof LuaClosure)) {
			LuaReturn retEnd = caller.protectedCall(thread, objEnd);
			if(!retEnd.isSuccess()) {
				System.err.println("End() function failed:");
				showLuaError(retEnd);
			} else {
				System.out.println("End() OK");
			}
		}
	}

	@Override
	public void PrepareThink() {
		super.PrepareThink();

		Object objPrepareThink = closure.env.rawget("PrepareThink");
		if((objPrepareThink != null) && (objPrepareThink instanceof LuaClosure)) {
			LuaReturn ret = caller.protectedCall(thread, objPrepareThink);
			if(!ret.isSuccess()) {
				System.err.println("PrepareThink() function failed:");
				showLuaError(ret);
			} else {
				//System.out.println("PrepareThink() OK");
			}
		}
	}

	@Override
	public int Evalfield(int depth) {
		Object objEvalfield = closure.env.rawget("EvalField");
		if((objEvalfield != null) && (objEvalfield instanceof LuaClosure)) {
			LuaReturn ret = caller.protectedCall(thread, objEvalfield, depth);
			if(!ret.isSuccess()) {
				System.err.println("EvalField(" + depth + ") function failed:");
				showLuaError(ret);
				return EVAL_MIN;
			} else {
				try {
					int score = converterManager.fromLuaToJava(ret.getFirst(), int.class);
					//System.out.println("EvalField(" + depth + ") :" + score);
					return score;
				} catch (Exception e) {
					System.err.println("EvalField(" + depth + ") function returned non integer-value");
					e.printStackTrace();
					return EVAL_MIN;
				}
			}
		}
		return EvalfieldRange(depth, 0, width_);
	}

	@Override
	public int EvalfieldRange(int depth, int left, int right) {
		Object objEvalfieldRange = closure.env.rawget("EvalFieldRange");
		if((objEvalfieldRange != null) && (objEvalfieldRange instanceof LuaClosure)) {
			LuaReturn ret = caller.protectedCall(thread, objEvalfieldRange, depth, left, right);
			if(!ret.isSuccess()) {
				System.err.println("EvalFieldRange(" + depth + "," + left + "," + right + ") function failed:");
				showLuaError(ret);
				return EVAL_MIN;
			} else {
				try {
					int score = converterManager.fromLuaToJava(ret.getFirst(), int.class);
					//System.out.println("EvalFieldRange(" + depth + "," + left + "," + right + ") :" + score);
					return score;
				} catch (Exception e) {
					System.err.println("EvalFieldRange(" + depth + "," + left + "," + right + ") function returned non integer-value");
					e.printStackTrace();
					return EVAL_MIN;
				}
			}
		}
		return EVAL_MIN;
	}

	@LuaMethod(name = "SetThinkDepth", global = true)
	public void luaSetThinkDepth(int depth) {
		System.out.println("depth = " + depth);
	}

	@LuaMethod(name = "SetThinkDelayFrame", global = true)
	public void luaSetThinkDelayFrame(int thinkDelay) {
		reqThinkDelayFrame_ = thinkDelay;
		System.out.println("reqThinkDelayFrame_ = " + reqThinkDelayFrame_);
	}

	@LuaMethod(name = "SetMoveDelayFrame", global = true)
	public void luaSetMoveDelayFrame(int moveDelay) {
		reqMoveDelayFrame_ = moveDelay;
		System.out.println("reqMoveDelayFrame_ = " + reqMoveDelayFrame_);
	}

	@LuaMethod(name = "SetHoldMinDiff", global = true)
	public void luaSetHoldMinDiff(int holdMinDiff) {
		holdMinDiff_ = holdMinDiff;
		System.out.println("holdMinDiff_ = " + holdMinDiff_);
	}

	@LuaMethod(name = "SetTSpinMinDiff", global = true)
	public void luaSetTSpinMinDiff(int tspinMinDiff) {
		tspinMinDiff_ = tspinMinDiff;
		System.out.println("tspinMinDiff_ = " + tspinMinDiff_);
	}

	@LuaMethod(name = "SetEnableTSpin", global = true)
	public void luaSetEnableTSpin(int enable) {
		enableTSpin_ = (enable == 1);
		System.out.println("enableTSpin_ = " + enableTSpin_);
	}

	@LuaMethod(name = "GetWidth", global = true)
	public int luaGetWidth() {
		return width_;
	}

	@LuaMethod(name = "GetHeight", global = true)
	public int luaGetHeight() {
		return height_;
	}

	@LuaMethod(name = "GetBlockLastHeight", global = true)
	public int luaGetBlockLastHeight() {
		return curBlockLastHeight_;
	}

	@LuaMethod(name = "GetBlockHighest", global = true)
	public int luaGetBlockHighest() {
		return curBlockHighest_;
	}

	@LuaMethod(name = "GetComboCount", global = true)
	public int luaGetComboCount() {
		return curComboCnt_;
	}

	@LuaMethod(name = "GetField", global = true)
	public int luaGetField(int x, int y) {
		return GetField(x, y) ? 1 : 0;
	}

	@LuaMethod(name = "Debug", global = true)
	public void luaDebug(String str) {
		System.out.println("Lua Debug:" + str);
	}
}
