package tk.to_p.blankedoutbox.garlicclone.core;

/**
 * Constants of Garlic engine
 */
public class GarlicConst {
	/** Tetrimino ID */
	public static final int TetriminoType_None = -1,
							TetriminoType_O = 0,
							TetriminoType_I = 1,
							TetriminoType_T = 2,
							TetriminoType_L = 3,
							TetriminoType_J = 4,
							TetriminoType_S = 5,
							TetriminoType_Z = 6,
							TetriminoType_Count = 7;

	/** Tetrimino Faces */
	public static final int FaceType_North = 0,
							FaceType_East = 1,
							FaceType_South = 2,
							FaceType_West = 3,
							FaceType_Count = 4;

	/** Tetrimino connection bit flags */
	public static final int StickyFlag_N = 1,
							StickyFlag_W = 2,
							StickyFlag_E = 4,
							StickyFlag_S = 8;

	/** Block colors */
	public static final int MinoColor_None = 0,
							MinoColor_Yellow = 1,
							MinoColor_LightBlue = 2,
							MinoColor_Purple = 3,
							MinoColor_Orange = 4,
							MinoColor_DarkBlue = 5,
							MinoColor_Green = 6,
							MinoColor_Red = 7,
							MinoColor_Garbage = 8,
							MinoColor_Concrete = 9,
							MinoColor_Count = 10;

	/** Ghost colors */
	public static final int MinoColor_Yellow_Ghost = 10,
							MinoColor_LightBlue_Ghost = 11,
							MinoColor_Purple_Ghost = 12,
							MinoColor_Orange_Ghost = 13,
							MinoColor_DarkBlue_Ghost = 14,
							MinoColor_Green_Ghost = 15,
							MinoColor_Red_Ghost = 16;

	/** Controller button bit flags */
	public static final int ControllerFlag_LEFT = 1,
							ControllerFlag_RIGHT = 2,
							ControllerFlag_CW = 4,
							ControllerFlag_CCW = 8,
							ControllerFlag_SOFT = 16,
							ControllerFlag_HARD = 32,
							ControllerFlag_HOLD = 64;

	/** Controller button IDs */
	public static final int BTN_LEFT = 0,
							BTN_RIGHT = 1,
							BTN_CW = 2,
							BTN_CCW = 3,
							BTN_SOFT = 4,
							BTN_HARD = 5,
							BTN_HOLD = 6,
							BTN_COUNT = 7;

	/** Cause of death IDs (Not really used) */
	public static final int GameOver_TopOut = 1,
							GameOver_LockOut = 2,
							GameOver_BlockOut = 3;

	/** Number of blocks in a Tetrimino */
	public static final int Tetri = 4;

	/** Tetrimino Objects */
	public static TetriMinos[] TetriMinos_;

	/**
	 * [Added] Tetrimino rotation patterns. Please note that the O piece does not rotate at all!
	 */
	private static final int[][][][] TETRIMINO_POS =
	{
		{ {{1, 0}, {2, 0}, {1, 1}, {2, 1}}, {{1, 0}, {2, 0}, {1, 1}, {2, 1}}, {{1, 0}, {2, 0}, {1, 1}, {2, 1}}, {{1, 0}, {2, 0}, {1, 1}, {2, 1}} }, // O
		{ {{0, 1}, {1, 1}, {2, 1}, {3, 1}}, {{2, 0}, {2, 1}, {2, 2}, {2, 3}}, {{0, 2}, {1, 2}, {2, 2}, {3, 2}}, {{1, 0}, {1, 1}, {1, 2}, {1, 3}} }, // I
		{ {{1, 0}, {0, 1}, {1, 1}, {2, 1}}, {{1, 0}, {1, 1}, {1, 2}, {2, 1}}, {{1, 2}, {0, 1}, {1, 1}, {2, 1}}, {{1, 0}, {1, 1}, {1, 2}, {0, 1}} }, // T
		{ {{2, 0}, {0, 1}, {1, 1}, {2, 1}}, {{1, 0}, {1, 1}, {1, 2}, {2, 2}}, {{0, 2}, {0, 1}, {1, 1}, {2, 1}}, {{1, 0}, {1, 1}, {1, 2}, {0, 0}} }, // L
		{ {{0, 0}, {0, 1}, {1, 1}, {2, 1}}, {{1, 0}, {1, 1}, {1, 2}, {2, 0}}, {{2, 2}, {0, 1}, {1, 1}, {2, 1}}, {{1, 0}, {1, 1}, {1, 2}, {0, 2}} }, // J
		{ {{1, 0}, {2, 0}, {0, 1}, {1, 1}}, {{1, 0}, {1, 1}, {2, 1}, {2, 2}}, {{0, 2}, {1, 2}, {1, 1}, {2, 1}}, {{0, 0}, {0, 1}, {1, 1}, {1, 2}} }, // S
		{ {{0, 0}, {1, 0}, {1, 1}, {2, 1}}, {{1, 1}, {1, 2}, {2, 0}, {2, 1}}, {{0, 1}, {1, 1}, {1, 2}, {2, 2}}, {{0, 1}, {0, 2}, {1, 0}, {1, 1}} }, // Z
	};

	/**
	 * Initialize Tetrimino objects
	 */
	public static void InitGarlic() {
		TetriMinos_ = new TetriMinos[TetriminoType_Count];

		TetriMinos_[TetriminoType_O] = new TetriMinos(TetriminoType_O);
		Make4X4Block(TetriMinos_[TetriminoType_O].tetriMinos_[FaceType_North], TETRIMINO_POS[TetriminoType_O][FaceType_North], MinoColor_Yellow);
		Make4X4Block(TetriMinos_[TetriminoType_O].tetriMinos_[FaceType_East],  TETRIMINO_POS[TetriminoType_O][FaceType_East],  MinoColor_Yellow);
		Make4X4Block(TetriMinos_[TetriminoType_O].tetriMinos_[FaceType_South], TETRIMINO_POS[TetriminoType_O][FaceType_South], MinoColor_Yellow);
		Make4X4Block(TetriMinos_[TetriminoType_O].tetriMinos_[FaceType_West],  TETRIMINO_POS[TetriminoType_O][FaceType_West],  MinoColor_Yellow);
		TetriMinos_[TetriminoType_I] = new TetriMinos(TetriminoType_I);
		Make4X4Block(TetriMinos_[TetriminoType_I].tetriMinos_[FaceType_North], TETRIMINO_POS[TetriminoType_I][FaceType_North], MinoColor_LightBlue);
		Make4X4Block(TetriMinos_[TetriminoType_I].tetriMinos_[FaceType_East],  TETRIMINO_POS[TetriminoType_I][FaceType_East],  MinoColor_LightBlue);
		Make4X4Block(TetriMinos_[TetriminoType_I].tetriMinos_[FaceType_South], TETRIMINO_POS[TetriminoType_I][FaceType_South], MinoColor_LightBlue);
		Make4X4Block(TetriMinos_[TetriminoType_I].tetriMinos_[FaceType_West],  TETRIMINO_POS[TetriminoType_I][FaceType_West],  MinoColor_LightBlue);
		TetriMinos_[TetriminoType_T] = new TetriMinos(TetriminoType_T);
		Make4X4Block(TetriMinos_[TetriminoType_T].tetriMinos_[FaceType_North], TETRIMINO_POS[TetriminoType_T][FaceType_North], MinoColor_Purple);
		Make4X4Block(TetriMinos_[TetriminoType_T].tetriMinos_[FaceType_East],  TETRIMINO_POS[TetriminoType_T][FaceType_East],  MinoColor_Purple);
		Make4X4Block(TetriMinos_[TetriminoType_T].tetriMinos_[FaceType_South], TETRIMINO_POS[TetriminoType_T][FaceType_South], MinoColor_Purple);
		Make4X4Block(TetriMinos_[TetriminoType_T].tetriMinos_[FaceType_West],  TETRIMINO_POS[TetriminoType_T][FaceType_West],  MinoColor_Purple);
		TetriMinos_[TetriminoType_L] = new TetriMinos(TetriminoType_L);
		Make4X4Block(TetriMinos_[TetriminoType_L].tetriMinos_[FaceType_North], TETRIMINO_POS[TetriminoType_L][FaceType_North], MinoColor_Orange);
		Make4X4Block(TetriMinos_[TetriminoType_L].tetriMinos_[FaceType_East],  TETRIMINO_POS[TetriminoType_L][FaceType_East],  MinoColor_Orange);
		Make4X4Block(TetriMinos_[TetriminoType_L].tetriMinos_[FaceType_South], TETRIMINO_POS[TetriminoType_L][FaceType_South], MinoColor_Orange);
		Make4X4Block(TetriMinos_[TetriminoType_L].tetriMinos_[FaceType_West],  TETRIMINO_POS[TetriminoType_L][FaceType_West],  MinoColor_Orange);
		TetriMinos_[TetriminoType_J] = new TetriMinos(TetriminoType_J);
		Make4X4Block(TetriMinos_[TetriminoType_J].tetriMinos_[FaceType_North], TETRIMINO_POS[TetriminoType_J][FaceType_North], MinoColor_DarkBlue);
		Make4X4Block(TetriMinos_[TetriminoType_J].tetriMinos_[FaceType_East],  TETRIMINO_POS[TetriminoType_J][FaceType_East],  MinoColor_DarkBlue);
		Make4X4Block(TetriMinos_[TetriminoType_J].tetriMinos_[FaceType_South], TETRIMINO_POS[TetriminoType_J][FaceType_South], MinoColor_DarkBlue);
		Make4X4Block(TetriMinos_[TetriminoType_J].tetriMinos_[FaceType_West],  TETRIMINO_POS[TetriminoType_J][FaceType_West],  MinoColor_DarkBlue);
		TetriMinos_[TetriminoType_S] = new TetriMinos(TetriminoType_S);
		Make4X4Block(TetriMinos_[TetriminoType_S].tetriMinos_[FaceType_North], TETRIMINO_POS[TetriminoType_S][FaceType_North], MinoColor_Green);
		Make4X4Block(TetriMinos_[TetriminoType_S].tetriMinos_[FaceType_East],  TETRIMINO_POS[TetriminoType_S][FaceType_East],  MinoColor_Green);
		Make4X4Block(TetriMinos_[TetriminoType_S].tetriMinos_[FaceType_South], TETRIMINO_POS[TetriminoType_S][FaceType_South], MinoColor_Green);
		Make4X4Block(TetriMinos_[TetriminoType_S].tetriMinos_[FaceType_West],  TETRIMINO_POS[TetriminoType_S][FaceType_West],  MinoColor_Green);
		TetriMinos_[TetriminoType_Z] = new TetriMinos(TetriminoType_Z);
		Make4X4Block(TetriMinos_[TetriminoType_Z].tetriMinos_[FaceType_North], TETRIMINO_POS[TetriminoType_Z][FaceType_North], MinoColor_Red);
		Make4X4Block(TetriMinos_[TetriminoType_Z].tetriMinos_[FaceType_East],  TETRIMINO_POS[TetriminoType_Z][FaceType_East],  MinoColor_Red);
		Make4X4Block(TetriMinos_[TetriminoType_Z].tetriMinos_[FaceType_South], TETRIMINO_POS[TetriminoType_Z][FaceType_South], MinoColor_Red);
		Make4X4Block(TetriMinos_[TetriminoType_Z].tetriMinos_[FaceType_West],  TETRIMINO_POS[TetriminoType_Z][FaceType_West],  MinoColor_Red);
	}

	/**
	 * Create a Tetrimino object
	 * @param tetriMino Tetrimino object
	 * @param pos Face type
	 * @param clr Color
	 */
	private static void Make4X4Block(TetriMinos.TetriMino tetriMino, int[][] pos, int clr) {
		tetriMino.xMin_ = pos[0][0];
		tetriMino.xMax_ = pos[0][0];

		for(int i = 0; i < Tetri; i++) {
			tetriMino.minos_[i].x_ = pos[i][0];
			tetriMino.minos_[i].y_ = pos[i][1];
			tetriMino.minos_[i].color_ = clr;
			tetriMino.xMin_ = Math.min(tetriMino.xMin_, tetriMino.minos_[i].x_);
			tetriMino.xMax_ = Math.max(tetriMino.xMax_, tetriMino.minos_[i].x_);
		}

		for(int i = 0; i < Tetri; i++) {
			int x = pos[i][0];
			int y = pos[i][1];
			int stickyFlag = 0;

			if(Make4X4Block_Find(tetriMino, x, y-1)) stickyFlag |= StickyFlag_N;
			if(Make4X4Block_Find(tetriMino, x+1, y)) stickyFlag |= StickyFlag_E;
			if(Make4X4Block_Find(tetriMino, x, y+1)) stickyFlag |= StickyFlag_S;
			if(Make4X4Block_Find(tetriMino, x-1, y)) stickyFlag |= StickyFlag_W;

			tetriMino.minos_[i].sticky_ = stickyFlag;
		}
	}

	/**
	 * Returns if there are connection(s) to the specified position.<br>
	 * [Became independent] It was an unnamed, nested subroutine of Make4X4Block.
	 * @param tetriMino [Added] Tetrimino
	 * @param x X positon
	 * @param y Y positon
	 * @return Returns <code>true</code> if there are connection(s) to the specified position
	 */
	private static boolean Make4X4Block_Find(TetriMinos.TetriMino tetriMino, int x, int y) {
		for(int i = 0; i < Tetri; i++) {
			if(tetriMino.minos_[i].x_ == x && tetriMino.minos_[i].y_ == y) {
				return true;
			}
		}
		return false;
	}
}
