package tk.to_p.blankedoutbox.garlicclone.core;
import java.util.Vector;

/**
 * Piece sequence generator (the "Random Generator")
 */
public class PlayerBag {
	/** Player that owns this class */
	public Player player_;

	/** Random number generator */
	public RandomStream randomStream_ = new RandomStream();

	/** Tetrimino Bag */
	public Vector<Integer> tetriMinosInBag_ = new Vector<Integer>();

	/**
	 * Constructor
	 * @param player Player object that owns this PlayerBag
	 */
	public PlayerBag(Player player) {
		player_ = player;
	}

	/**
	 * Initialize
	 * @param seed Random-Seed
	 */
	public void Init(long seed) {
		randomStream_.SetSeed(seed);
	}

	/**
	 * Pop the first next piece (It will remove the piece from the bag)
	 * @return Piece ID
	 */
	public int Pop() {
		int next = GetNext(0);
		tetriMinosInBag_.remove(0);
		return next;
	}

	/**
	 * Get the next piece (Does not remove)
	 * @param idx Piece Index
	 * @return Piece ID
	 * @throws ArrayIndexOutOfBoundsException If idx is less than 0 or bigger than the size of the bag
	 */
	public int GetNext(int idx) {
		CheckNextBag();
		return tetriMinosInBag_.get(idx);
	}

	/**
	 * Create the piece sequences
	 */
	public void CheckNextBag() {
		int bagSize = player_.gamePlay_.playerSetting_.BagSize;
		int bagSize2 = bagSize * 3;

		while(true) {
			if( !((tetriMinosInBag_.size() < bagSize2) && (tetriMinosInBag_.size() % bagSize == 0)) ) {
				break;
			}

			Vector<Integer> elements = new Vector<Integer>();
			for(int i = 0; i < player_.gamePlay_.playerSetting_.BagElements.size(); i++) {
				elements.addElement(player_.gamePlay_.playerSetting_.BagElements.get(i));
			}
			//elements.addAll(player_.gamePlay_.playerSetting_.BagElements);

			Vector<Integer> tempArray = new Vector<Integer>();

			for(int i = 0; i < elements.size(); i++) {
				tempArray.addElement(randomStream_.RandomInt(0, 100000));
			}

			for(int i = 0; i < elements.size(); i++) {
				for(int j = i + 1; j < elements.size(); j++) {
					if(tempArray.get(i) > tempArray.get(j)) {
						int temp = tempArray.get(i);
						tempArray.set(i, tempArray.get(j));
						tempArray.set(j, temp);
						temp = elements.get(i);
						elements.set(i, elements.get(j));
						elements.set(j, temp);
					}
				}
			}

			for(int i = 0; i < bagSize; i++) {
				int tetriminoType = GarlicConst.TetriminoType_None;
				if(elements.size() > i) {
					tetriminoType = elements.get(i);
				}
				if(tetriminoType == GarlicConst.TetriminoType_None) {
					tetriminoType = randomStream_.RandomInt(GarlicConst.TetriminoType_O, GarlicConst.TetriminoType_Z);
				}
				tetriMinosInBag_.addElement(tetriminoType);
			}
		}
	}
}
